from fastapi import FastAPI, Depends
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller
from core.error import HTTPAPIException
from module.auth.core import Authorizer
from module.snake_module_name.schema.snake_entity_name import (
    PascalEntityName, PascalEntityNameData, PascalEntityNameResult
)
from module.auth.schema.token import TokenData
from module.auth.component import token_scheme


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register API for "snake_module_name.snake_entity_name"')

    @app.get(
        '/api/v1/kebab-module-name/kebab-plural-entity-name', response_model=PascalEntityNameResult
    )
    async def get_snake_entity_names(
        keyword: str = '', limit: int = 100, offset: int = 0,
        user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'snake_module_name:snake_entity_name:get'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'snake_module_name_get_snake_entity_name',
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.dict()
            )
            return PascalEntityNameResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get(
        '/api/v1/kebab-module-name/kebab-plural-entity-name/{id}', response_model=PascalEntityName
    )
    async def get_snake_entity_name_by_id(
        id: str, user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'snake_module_name:snake_entity_name:get_by_id'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'snake_module_name_get_snake_entity_name_by_id',
                id=id, user_token_data=user_token_data.dict()
            )
            return PascalEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post(
        '/api/v1/kebab-module-name/kebab-plural-entity-name', response_model=PascalEntityName
    )
    async def insert_snake_entity_name(
        data: PascalEntityNameData,
        user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'snake_module_name:snake_entity_name:insert'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'snake_module_name_insert_snake_entity_name',
                data=data.dict(), user_token_data=user_token_data.dict()
            )
            return PascalEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put(
        '/api/v1/kebab-module-name/kebab-plural-entity-name/{id}', response_model=PascalEntityName
    )
    async def update_snake_entity_name(
        id: str, data: PascalEntityNameData,
        user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'snake_module_name:snake_entity_name:update'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'snake_module_name_update_snake_entity_name',
                id=id, data=data.dict(), user_token_data=user_token_data.dict()
            )
            return PascalEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete(
        '/api/v1/kebab-module-name/kebab-plural-entity-name/{id}', response_model=PascalEntityName
    )
    async def delete_snake_entity_name(
        id: str, user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'snake_module_name:snake_entity_name:delete'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'snake_module_name_delete_snake_entity_name',
                id=id, user_token_data=user_token_data.dict()
            )
            return PascalEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from core.model import RepoModel
from module.snake_module_name.schema.snake_entity_name import (
    PascalEntityName, PascalEntityNameData, PascalEntityNameResult
)


class PascalEntityNameModel(
    RepoModel[PascalEntityName, PascalEntityNameData, PascalEntityNameResult]
):
    schema_result_cls = PascalEntityNameResult



from abc import ABC, abstractmethod
from module.auth.schema.token import TokenData


class TokenUtil(ABC):

    @abstractmethod
    def encode(self, token_data: TokenData) -> str:
        pass

    @abstractmethod
    def decode(self, token_str: str) -> TokenData:
        pass



from module.auth.core import TokenScheme, create_oauth2_bearer_token_scheme
from module.auth.component.token_util import token_util
from module.auth.component.user import guest_user


token_scheme: TokenScheme = create_oauth2_bearer_token_scheme(
    guest_user=guest_user,
    token_util=token_util,
    token_url='/api/v1/auth/login-oauth',
    token_cookie_key='auth_token'
)



from config import app_auth_token_expire_seconds, app_auth_admin_active
from module.auth.component.repo.user_repo import user_repo
from module.auth.entity.user.model import UserModel
from module.auth.component import token_util
from module.auth.component.user import (
    admin_user, admin_user_password, guest_user
)
from module.auth.component.model.permission_model import permission_model


user_model: UserModel = UserModel(
    repo=user_repo,
    permission_model=permission_model,
    token_util=token_util,
    expire_seconds=app_auth_token_expire_seconds,
    guest_user=guest_user,
    admin_user=admin_user if app_auth_admin_active else None,
    admin_user_password=admin_user_password
)



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....runner import runner
from .._common.input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
    column_name_input, column_type_input
)
from .._common.helper import validate_project_dir
from ....helper import util
from ....helper.codemod.add_property_to_class import add_property_to_class
from ....helper.codemod.add_key_value_to_dict import add_key_value_to_dict
from ....helper.file.text import read_text_file_async, write_text_file_async

import asyncio
import os

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[
        project_dir_input, app_name_input, module_name_input, entity_name_input
    ],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    source_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(source_dir):
        raise Exception(f'Source does not exist: {source_dir}')
    module_path = os.path.join(source_dir, 'module', snake_module_name)
    if os.path.exists(module_path):
        raise Exception(f'Module directory already exists: {module_path}')
    entity_path = os.path.join(module_path, 'entity', snake_entity_name)
    if os.path.exists(entity_path):
        raise Exception(f'Entity directory already exists: {entity_path}')


@python_task(
    name='fastapp-field',
    group=project_add_group,
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        column_name_input,
        column_type_input,
    ],
    upstreams=[
        validate,
    ],
    runner=runner
)
async def add_fastapp_field(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    column_name = kwargs.get('column_name')
    column_type = kwargs.get('column_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    pascal_entity_name = util.to_pascal_case(entity_name)
    snake_column_name = util.to_snake_case(column_name)
    await asyncio.gather(
        asyncio.create_task(add_column_to_test(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, snake_column_name, column_type
        )),
        asyncio.create_task(add_column_to_schema(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
        asyncio.create_task(add_column_to_repo(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
    )
    task.print_out('Success')


async def add_column_to_test(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    test_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'test', snake_module_name,
        f'test_{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {test_file_path}')
    code = await read_text_file_async(test_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the test'
    )
    dict_names = [
        'inserted_success_data', 'to_be_updated_success_data',
        'updated_success_data', 'to_be_deleted_success_data'
    ]
    for dict_name in dict_names:
        code = add_key_value_to_dict(
            code=code,
            dict_name=dict_name,
            key=f"'{snake_column_name}'",
            value="''"
        )
    task.print_out(f'Write modified code to: {test_file_path}')
    await write_text_file_async(test_file_path, code)


async def add_column_to_schema(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    schema_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'schema', f'{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {schema_file_path}')
    code = await read_text_file_async(schema_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the schema'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'{pascal_entity_name}Data',
        property_name=snake_column_name,
        property_type='str'
    )
    task.print_out(f'Write modified code to: {schema_file_path}')
    await write_text_file_async(schema_file_path, code)


async def add_column_to_repo(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    repo_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'entity', snake_entity_name, 'repo.py'
    )
    task.print_out(f'Read code from: {repo_file_path}')
    code = await read_text_file_async(repo_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the repo'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'DBEntity{pascal_entity_name}',
        property_name=snake_column_name,
        property_type='Column',
        property_value='Column(String)'
    )
    task.print_out(f'Write modified code to: {repo_file_path}')
    await write_text_file_async(repo_file_path, code)



from module.auth.core.authorizer.authorizer import Authorizer
from module.auth.core.authorizer.rpc_authorizer import RPCAuthorizer
from module.auth.core.password_hasher.password_hasher import PasswordHasher
from module.auth.core.password_hasher.bcrypt_password_hasher import (
    BcryptPasswordHasher
)
from module.auth.core.token_scheme.token_sheme import TokenScheme
from module.auth.core.token_scheme.oauth2_bearer_token_scheme import (
    create_oauth2_bearer_token_scheme
)
from module.auth.core.token_util.token_util import TokenUtil
from module.auth.core.token_util.jwt_token_util import JWTTokenUtil

assert Authorizer
assert RPCAuthorizer
assert PasswordHasher
assert BcryptPasswordHasher
assert TokenScheme
assert create_oauth2_bearer_token_scheme
assert TokenUtil
assert JWTTokenUtil



from core.model import RepoModel
from module.auth.schema.permission import (
    Permission, PermissionData, PermissionResult
)
from module.auth.entity.permission.repo import PermissionRepo


class PermissionModel(
    RepoModel[Permission, PermissionData, PermissionResult]
):
    schema_result_cls = PermissionResult

    def __init__(self, repo: PermissionRepo):
        self.repo = repo

    async def ensure_permission(self, data: PermissionData):
        try:
            await self.repo.get_by_name(data.name)
        except Exception as e:
            error_message = f'{e}'
            if error_message.lower().startswith('not found'):
                await self.repo.insert(data)
                return
            raise e



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task_input.str_input import StrInput
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common.input import (
   project_dir_input, task_name_input, http_port_input, env_prefix_input
)
from .._common.helper import validate_project_dir, create_register_task_module
from .._common.lock import new_task_lock
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[
        project_dir_input,
        task_name_input
    ],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    task_name = kwargs.get('task_name')
    automation_file = os.path.join(
        project_dir, '_automate', f'{util.to_snake_case(task_name)}.py'
    )
    if os.path.exists(automation_file):
        raise Exception(f'Automation file already exists: {automation_file}')
    source_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(task_name)}'
    )
    if os.path.exists(source_dir):
        raise Exception(f'Source already exists: {source_dir}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        task_name_input,
        http_port_input,
        StrInput(
            name='compose-command',
            description='Compose command (e.g., up, down, start, remove, etc)',
            prompt='Compose command (e.g., up, down, start, remove, etc)',
            default='up'
        ),
        env_prefix_input,
    ],
    upstreams=[validate],
    replacements={
        'taskName': '{{input.task_name}}',
        'httpPort': '{{util.coalesce(input.http_port, "3000")}}',
        'composeCommand': '{{ util.coalesce(input.compose_command, "up") }}',
        'ENV_PREFIX': '{{ util.coalesce(input.env_prefix, "MY").upper() }}',
    },
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    locks=[new_task_lock]
)

register_task_module = create_register_task_module(
    upstreams=[copy_resource]
)


@python_task(
    name='docker-compose-task',
    group=project_add_group,
    upstreams=[register_task_module],
    runner=runner
)
async def add_docker_compose_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from typing import List, Optional
from module.auth.schema.permission import PermissionData
from module.auth.component.model.permission_model import permission_model


async def ensure_entity_permission(
    module_name: str,
    entity_name: str,
    action_names: Optional[List[str]] = None
):
    if action_names is None:
        action_names = ['get', 'get_by_id', 'insert', 'update', 'delete']
    for action_name in action_names:
        await permission_model.ensure_permission(PermissionData(
            name=f'{module_name}:{entity_name}:{action_name}',
            description=f'{module_name}:{entity_name}:{action_name}'
        ))


async def register_permission():
    await ensure_entity_permission(
        module_name='auth', entity_name='group'
    )
    await ensure_entity_permission(
        module_name='auth', entity_name='permission'
    )
    await ensure_entity_permission(
        module_name='auth', entity_name='user'
    )



from module.auth.component.authorizer import authorizer
from module.auth.component.base import Base
from module.auth.component.password_hasher import password_hasher
from module.auth.component.token_scheme import token_scheme
from module.auth.component.token_util import token_util
from module.auth.component.user import (
    admin_user, admin_user_password, guest_user
)

assert authorizer
assert Base
assert password_hasher
assert token_scheme
assert token_util
assert admin_user
assert admin_user_password
assert guest_user



from typing import Optional
from starlette.requests import Request
from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from module.auth.core.token_util.token_util import TokenUtil
from module.auth.core.token_scheme.token_sheme import TokenScheme
from module.auth.schema.token import TokenData
from module.auth.schema.user import User


def create_oauth2_bearer_token_scheme(
    guest_user: User,
    token_util: TokenUtil,
    token_url: str,
    token_cookie_key: str
) -> TokenScheme:

    oauth2_scheme = OAuth2PasswordBearer(
        tokenUrl=token_url, auto_error=False
    )

    async def oauth2_bearer_token_scheme(
        request: Request,
        token: Optional[str] = Depends(oauth2_scheme)
    ) -> TokenData:
        token: Optional[str] = await oauth2_scheme(request)
        if token is None:
            request.cookies.get(token_cookie_key, None)
        if token is None:
            return TokenData(
                user_id=guest_user.id,
                username=guest_user.username,
                permission_names=[],
                expire_seconds=300
            )
        return token_util.decode(token)

    return oauth2_bearer_token_scheme



from core.model import RepoModel
from module.auth.schema.group import (
    Group, GroupData, GroupResult
)


class GroupModel(
    RepoModel[Group, GroupData, GroupResult]
):
    schema_result_cls = GroupResult



from typing import (
    Any, Callable, Iterable, Optional, Union
)
from typeguard import typechecked
from ..task_input.base_input import BaseInput
from ..task_env.env import Env
from ..task_env.env_file import EnvFile
from ..task.base_task import Group
from ..action.runner import Runner
from .base_task import BaseTask
from .task import Task


@typechecked
def python_task(
    name: str,
    group: Optional[Group] = None,
    inputs: Iterable[BaseInput] = [],
    envs: Iterable[Env] = [],
    env_files: Iterable[EnvFile] = [],
    icon: Optional[str] = None,
    color: Optional[str] = None,
    description: str = '',
    upstreams: Iterable[BaseTask] = [],
    checkers: Iterable[BaseTask] = [],
    checking_interval: float = 0.1,
    retry: int = 2,
    retry_interval: float = 1,
    skip_execution: Union[str, bool] = False,
    runner: Optional[Runner] = None
) -> Callable[[Callable[..., Any]], Task]:
    def _create_task(fn: Callable[..., Any]) -> Task:
        task = Task(
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            run=fn,
            skip_execution=skip_execution
        )
        if runner is not None:
            runner.register(task)
        return task
    return _create_task
